;;;; Copyright (C) 2016 Lily Carpenter

;;;; This program is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU Affero General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU Affero General Public License for more details.

;;;; You should have received a copy of the GNU Affero General Public License
;;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;;; crypts-and-corpses.lisp

(in-package #:crypts-and-corpses)

(defstruct game-map
  (width  99 :type fixnum :read-only t)
  (height 39 :type fixnum :read-only t)
  (dungeon nil :read-only t)
  (mobiles (make-hash-table) :type hash-table :read-only t)
  (terrain (make-hash-table) :type hash-table :read-only t))

(defmethod initialize-instance :after ((map game-map) &rest initargs)
  (declare (ignore initargs))
  (with-slots (width height dungeon) map
    (setf dungeon (crawler:make-dungeon width height))))

(defclass entity ()
  ((width :initarg :width :type fixnum :initform 1)
   (height :initarg :height :type fixnum :initform 1)
   (location :initarg :location :initform 0 :type fixnum)
   (color :initarg :color :initform (color-from-name "#ffffff"))))

(defclass tile-entity (entity)
  ((codepoint :initarg :codepoint :initform #x3f :type fixnum)))

(defclass wall (tile-entity)
  ((codepoint :initarg :codepoint :initform #x23 :type fixnum)
   (color :initarg :color :initform (color-from-name "#999999"))))

(defclass floor-tile (tile-entity)
  ((codepoint :initarg :codepoint :initform #xb7 :type fixnum)
   (color :initarg :color :initform (color-from-name "#3f3f3f"))))

(defclass viewport (entity)
  ((width :initarg :width :type fixnum :initform 99)
   (height :initarg :height :type fixnum :initform 40)
   (location :initarg :location :type fixnum :initform 0)))

(defclass mobile (tile-entity)
  ((health :initarg :health :type fixnum :initform 10)
   (mana :initarg :mana :type fixnum :initform 0)
   (max-health :initarg :health :type fixnum :initform 10)
   (max-mana :initarg :mana :type fixnum :initform 10)
   (health-regen :initarg :health-regen :initform 0 :type fixnum)
   (mana-regen :initarg :mana-regen :initform 0 :type fixnum)))

(defclass player-character (mobile)
  ((health :initarg :health :type fixnum :initform 10)
   (mana :initarg :mana :type fixnum :initform 10)
   (codepoint :initform #x2639)
   (ranged-attack-damage :initarg :ranged-attack-damage :initform 5 :type fixnum)
   (melee-attack-damage :initarg :ranged-attack-damage :initform 2 :type fixnum)
   (ranged-attack-cost :initarg :ranged-attack-cost :initform 5 :type fixnum)
   (mana-regen :initarg :mana-regen :initform 3 :type fixnum)
   (color :initarg :color :initform (color-from-name "#800080"))))

(defclass enemy (mobile)
  ((codepoint :initform #x263a)
   (melee-attack-damage :initarg :ranged-attack-damage :initform 1 :type fixnum)
   (health-regen :initarg :health-regen :initform 1 :type fixnum)
   (color :initarg :color :initform (color-from-name "#1919ff"))))

(defclass minion (mobile)
  ((color :initarg :color :initform (color-from-name "yellow"))))

(defun calc-coord (location map)
  (values (mod location (game-map-width map)) (floor (/ location (game-map-width map)))))

(defun get-window-size ()
  (apply 'values (mapcar #'parse-integer (split-sequence #\x (terminal-get "window.size" "0x0")))))

(defun calc-location (x y map)
  (+ (* (game-map-width map) y) x))

(defgeneric set-location (object x y map))
(defmethod set-location ((object tile-entity) x y map)
  (setf (slot-value object 'location) (calc-location x y map)))

(defgeneric end-of-turn (object))

(defmethod end-of-turn ((object mobile))
  (incf (slot-value object 'health) (slot-value object 'health-regen))
  (incf (slot-value object 'mana) (slot-value object 'mana-regen))
  (when (> (slot-value object 'health) (slot-value object 'max-health))
    (setf (slot-value object 'health) (slot-value object 'max-health)))
  (when (> (slot-value object 'mana) (slot-value object 'max-mana))
    (setf (slot-value object 'mana) (slot-value object 'max-mana))))

(defgeneric move-left (object map))
(defgeneric move-right (object map))
(defgeneric move-up (object map))
(defgeneric move-down (object map))

(defgeneric update-location (object map old-location new-location))

(defun out-of-bounds-p (new-location map)
  (multiple-value-bind (x y) (calc-coord new-location map)
    (or (< x 0)
        (< y 0)
        (>= x (game-map-width map))
        (>= y (game-map-height map)))))

(defun overrun-p (old-location new-location map)
  (multiple-value-bind (old-x old-y) (calc-coord old-location map)
    (multiple-value-bind (new-x new-y) (calc-coord new-location map)
      (or (and (not (= old-y new-y))
               (not (= old-x new-x)))))))

(defun calculate-effective-location (object map location)
  (+
   location
   (1- (slot-value object 'width))
   (* (1- (slot-value object 'height)) (game-map-width map))))

(defun collision-p (new-location map)
  (let ((thing (gethash new-location (game-map-mobiles map)))
        (tile (gethash new-location (game-map-terrain map))))
    (or thing
        (and tile (not (typep tile 'floor-tile))))))

(defun bad-move-p (object map old-location new-location)
  (let ((old-effective-location (calculate-effective-location object map old-location))
        (new-effective-location (calculate-effective-location object map new-location)))
    (or (out-of-bounds-p new-effective-location map)
        (collision-p new-effective-location map)
        (overrun-p old-effective-location new-effective-location map))))

(defmethod update-location ((object entity) map old-location new-location)
  (unless (bad-move-p object map old-location new-location)
    (setf (slot-value object 'location) new-location)))

(defmethod update-location ((object mobile) map old-location new-location)
  (unless (bad-move-p object map old-location new-location)
    (setf (slot-value object 'location) new-location)
    (setf (gethash new-location (game-map-mobiles map)) object)
    (remhash old-location (game-map-mobiles map))))

(defmethod move-left ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (1- (slot-value object 'location))))
    (update-location object map old-location new-location)))

(defmethod move-up ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (- (slot-value object 'location) (game-map-width map))))
    (update-location object map old-location new-location)))

(defmethod move-down ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (+ (slot-value object 'location) (game-map-width map))))
    (update-location object map old-location new-location)))

(defmethod move-right ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (1+ (slot-value object 'location))))
    (update-location object map old-location new-location)))

(defun find-mobile-left (location map)
  (let ((found-object nil))
    (loop for i from (1- location) downto 0 do
      (setf found-object (gethash i (game-map-mobiles map)))
      (when (not (typep found-object 'mobile))
        (setf found-object nil))
          until found-object)
    found-object))

(defun find-mobile-right (location map)
  (let ((found-object nil))
    (loop for i from (1+ location) to (* (game-map-width map) (game-map-height map)) do
      (setf found-object (gethash i (game-map-mobiles map)))
      (when (not (typep found-object 'mobile))
        (setf found-object nil))
          until found-object)
    found-object))

(defun find-mobile-up (location map)
  (let ((found-object nil))
    (loop for i from (- location (game-map-width map)) downto 0 by (game-map-width map) do
      (setf found-object (gethash i (game-map-mobiles map)))
      (when (not (typep found-object 'mobile))
        (setf found-object nil))
          until found-object)
    found-object))

(defun find-mobile-down (location map)
  (let ((found-object nil))
    (loop for i from (+ location (game-map-width map)) to (* (game-map-width map) (game-map-height map)) by (game-map-width map) do
      (setf found-object (gethash i (game-map-mobiles map)))
      (when (not (typep found-object 'mobile))
        (setf found-object nil))
          until found-object)
    found-object))

(defgeneric shoot-left (object map))
(defgeneric shoot-right (object map))
(defgeneric shoot-up (object map))
(defgeneric shoot-down (object map))

(defmethod shoot-left ((object player-character) map)
  (let ((target (find-mobile-left (slot-value object 'location) map)))
    (when target
      (decf (slot-value target 'health) (slot-value object 'ranged-attack-damage))
      (decf (slot-value object 'mana) (slot-value object 'ranged-attack-cost)))))

(defmethod shoot-right ((object player-character) map)
  (let ((target (find-mobile-right (slot-value object 'location) map)))
    (when target
      (decf (slot-value target 'health) (slot-value object 'ranged-attack-damage))
      (decf (slot-value object 'mana) (slot-value object 'ranged-attack-cost)))))

(defmethod shoot-up ((object player-character) map)
  (let ((target (find-mobile-up (slot-value object 'location) map)))
    (when target
      (decf (slot-value target 'health) (slot-value object 'ranged-attack-damage))
      (decf (slot-value object 'mana) (slot-value object 'ranged-attack-cost)))))

(defmethod shoot-down ((object player-character) map)
  (let ((target (find-mobile-down (slot-value object 'location) map)))
    (when target
      (decf (slot-value target 'health) (slot-value object 'ranged-attack-damage))
      (decf (slot-value object 'mana) (slot-value object 'ranged-attack-cost)))))

(defgeneric bash-left (object map))
(defgeneric bash-right (object map))
(defgeneric bash-up (object map))
(defgeneric bash-down (object map))

(defmethod bash-left ((object player-character) map)
  (let* ((location (1- (slot-value object 'location)))
         (potential-target (gethash location (game-map-mobiles map)))
         (target (and potential-target (typep potential-target 'mobile))))
    (when target
      (decf (slot-value target 'health) (slot-value object 'melee-attack-damage)))))

(defmethod bash-right ((object player-character) map)
  (let* ((location (1+ (slot-value object 'location)))
         (potential-target (gethash location (game-map-mobiles map)))
         (target (and potential-target (typep potential-target 'mobile))))
    (when target
      (decf (slot-value target 'health) (slot-value object 'melee-attack-damage)))))

(defmethod bash-up ((object player-character) map)
  (let* ((location (- (slot-value object 'location) (game-map-width map)))
         (potential-target (gethash location (game-map-mobiles map)))
         (target (and potential-target (typep potential-target 'mobile))))
    (when target
      (decf (slot-value target 'health) (slot-value object 'melee-attack-damage)))))

(defmethod bash-down ((object player-character) map)
  (let* ((location (+ (slot-value object 'location) (game-map-width map)))
         (potential-target (gethash location (game-map-mobiles map)))
         (target (and potential-target (typep potential-target 'mobile))))
    (when target
      (decf (slot-value target 'health) (slot-value object 'melee-attack-damage)))))

(defun intro-message ()
  (multiple-value-bind (win-width win-height) (get-window-size)
    (terminal-print (floor (- (/ win-width 2) 10)) (floor (1- (/ win-height 2))) "Crypts and Corpses")
    (terminal-print (floor (- (/ win-width 2) 11)) (floor (1+ (/ win-height 2))) "Press enter to start"))
  (terminal-refresh))

(defun write-map-to-terminal (player map viewport)
  (let* ((player-location (slot-value player 'location))
         (current-room (find-room player-location map))
         (vision-range (if current-room
                           (with-slots ((x1 crawler::x1) (y1 crawler::y1) (x2 crawler::x2) (y2 crawler::y2)) current-room
                             (list (list (1- x1) (1- y1))
                                   (list x2 y2)))
                           (progn
                             (multiple-value-bind (x y) (calc-coord player-location map)
                               (list (list (1- x) (1- y))
                                     (list (1+ x) (1+ y))))))))
    (labels ((write-cell (key tile)
               (multiple-value-bind (x y) (calc-coord (- key (slot-value viewport 'location)) map)
                 (let ((x1 (caar vision-range))
                       (x2 (caadr vision-range))
                       (y1 (cadar vision-range))
                       (y2 (cadadr vision-range)))
                   (when (and (>= x x1)
                              (<= x x2)
                              (>= y y1)
                              (<= y y2))
                     (terminal-color (slot-value tile 'color))
                     (terminal-put x y (slot-value tile 'codepoint)))))))
      (maphash #'write-cell (game-map-terrain map))
      (maphash #'write-cell (game-map-mobiles map)))))

(defun nearest-open-tile (location map)
  (let ((left (1- location))
        (right (1+ location))
        (up (- location (game-map-width map)))
        (down (+ (game-map-width map) location))
        (mobiles (game-map-mobiles map)))
    (cond
      ((eq :not-found (gethash left mobiles :not-found)) left)
      ((eq :not-found (gethash up mobiles :not-found)) up)
      ((eq :not-found (gethash right mobiles :not-found)) right)
      ((eq :not-found (gethash down mobiles :not-found)) down))))

(defun in-room-p (location map room)
  (declare (type crawler::dungeon-room room))
  (multiple-value-bind (x y) (calc-coord location map)
    (with-slots ((x1 crawler::x1) (y1 crawler::y1) (x2 crawler::x2) (y2 crawler::y2)) room
      (and (>= x x1)
           (<= x x2)
           (>= y y1)
           (<= y y2)))))

(defun find-room (location map)
  (labels ((location-in-room-p (room) (in-room-p location map room)))
    (find-if #'location-in-room-p (slot-value (slot-value map 'dungeon) 'crawler::rooms))))

(defun find-arc (x y x0 y0 map)
  (labels ((calc-location-map (x y) (calc-location x y map)))
    (let ((coords (list (list (+ x x0) (+ y y0))
                        (list (+ y x0) (+ x y0))
                        (list (- x0 x) (+ y y0))
                        (list (- x0 y) (+ x y0))
                        (list (- x0 x) (- y0 y))
                        (list (- x0 y) (- y0 x))
                        (list (+ x x0) (- y0 y))
                        (list (+ y x0) (- y0 x)))))
      (loop for i in coords collect (apply #'calc-location-map i)))))

(defun find-circle (location radius map)
  (multiple-value-bind (x0 y0) (calc-coord location map)
    (let ((cur-radius radius)
          (y 0)
          (decision-over-2 (- 1 radius)))
      (loop while (<= y cur-radius)
            as coords = (find-arc cur-radius y x0 y0 map) do
              (incf y)
              (if (<= decision-over-2 0)
                  (incf decision-over-2 (1+ (* 2 y)))
                  (progn
                    (decf cur-radius)
                    (incf decision-over-2 (1+ (* 2 (- y cur-radius))))))
            collect coords))))

(defun find-line (location0 location1 map)
  (multiple-value-bind (x0 y0) (calc-coord location0 map)
    (multiple-value-bind (x1 y1) (calc-coord location1 map)
      (format t "~S ~S ~%" x0 x1)
      (when (< x1 x0)
        (return-from find-line (reverse (find-line location1 location0 map))))
      (let* ((delta-x (- x1 x0))
             (delta-y (- y1 y0))
             (decision (- (* 2 delta-y) delta-x))
             (y y0))
        (format t "~S ~S ~S ~%" delta-x delta-y decision)
        (when (> decision 0)
          (incf y)
          (decf decision (* 2 delta-x)))
        (loop for x from (+ x0 1) to x1
              as coord = (list x y) do
                (incf decision (* 2 delta-y))
                (when (> decision 0)
                  (incf y)
                  (decf decision (* 2 delta-x)))
              collect coord into coords
              finally (return coords))))))

(defun fill-map-from-crawler (map player)
  (with-slots (width height dungeon mobiles terrain) map
    (dotimes (x width)
      (dotimes (y height)
        (let ((tile (aref (crawler:tile-map dungeon) x y)))
          (cond
            ((not (crawler:walkablep tile))
             (let ((wall (make-instance 'wall)))
               (set-location wall x y map)
               (setf (gethash (calc-location x y map) terrain) wall)))
            ((eq (crawler:map-feature tile) :stairs-up)
             (set-location player x y map)
             (setf (gethash (calc-location x y map) mobiles) player)
             (let ((floor (make-instance 'floor-tile)))
               (set-location floor x y map)
               (setf (gethash (calc-location x y map) terrain) floor)))
            (:otherwise
             (let ((floor (make-instance 'floor-tile)))
               (set-location floor x y map)
               (setf (gethash (calc-location x y map) terrain) floor)))))))))

(defun initialize-map (map player)
  (fill-map-from-crawler map player)
  (setf (gethash (slot-value player 'location) (game-map-mobiles map)) player)
  (let* ((location (nearest-open-tile (slot-value player 'location) map))
         (enemy (make-instance 'enemy :location location)))
    (setf (gethash location (game-map-mobiles map)) enemy))
  map)

(defun handle-input (key map viewport player prev-key)
  (cond
    ;; Ranged attack
    ((equal prev-key bearlibterminal-ffi:+tk-r+)
     (cond
       ((= key bearlibterminal-ffi:+tk-up+) (shoot-up player map))
       ((= key bearlibterminal-ffi:+tk-down+) (shoot-down player map))
       ((= key bearlibterminal-ffi:+tk-right+) (shoot-right player map))
       ((= key bearlibterminal-ffi:+tk-left+) (shoot-left player map))
       ((= key bearlibterminal-ffi:+tk-enter+) t))) ; cancel
    ;; Melee attack
    ((equal prev-key bearlibterminal-ffi:+tk-m+)
     (cond
       ((= key bearlibterminal-ffi:+tk-up+) (bash-up player map))
       ((= key bearlibterminal-ffi:+tk-down+) (bash-down player map))
       ((= key bearlibterminal-ffi:+tk-right+) (bash-right player map))
       ((= key bearlibterminal-ffi:+tk-left+) (bash-left player map))
       ((= key bearlibterminal-ffi:+tk-enter+) t))) ; cancel
    ;; Not a stateful command
    (:otherwise
     (cond
       ((= key bearlibterminal-ffi:+tk-r+) t) ; Prepare for ranged
       ((= key bearlibterminal-ffi:+tk-m+) t) ; Prepare for melee
       ((= key bearlibterminal-ffi:+tk-enter+) t) ; idle
       ((= key bearlibterminal-ffi:+tk-up+) (move-up player map))
       ((= key bearlibterminal-ffi:+tk-down+) (move-down player map))
       ((= key bearlibterminal-ffi:+tk-right+) (move-right player map))
       ((= key bearlibterminal-ffi:+tk-left+) (move-left player map))
       ((= key bearlibterminal-ffi:+tk-w+) (move-up viewport map))
       ((= key bearlibterminal-ffi:+tk-a+) (move-left viewport map))
       ((= key bearlibterminal-ffi:+tk-d+) (move-right viewport map))
       ((= key bearlibterminal-ffi:+tk-s+) (move-down viewport map))
       ((= key bearlibterminal-ffi:+tk-space+) (update-location viewport map (slot-value viewport 'location) (slot-value player 'location)))))))

(defun create-viewport (map)
  (multiple-value-bind (width height) (get-window-size)
    (when (> width (game-map-width map))
      (setf width (game-map-width map)))
    (when (> height (game-map-height map))
      (setf height (game-map-height map)))
    (make-instance 'viewport :width width :height height)))

(defun end-of-turn-tasks (player map)
  (declare (ignore player))
  (maphash (lambda (location object)
             (when (typep object 'mobile)
               (if (>= 0 (slot-value object 'health))
                   (progn
                     (setf (slot-value object 'location) nil)
                     (remhash location (game-map-mobiles map)))
                   (end-of-turn object)))) (game-map-mobiles map)))

(defun write-status-to-terminal (player prev-key)
  (declare (ignore prev-key))
  (multiple-value-bind (width height) (get-window-size)
    (declare (ignore width))
    (terminal-color (color-from-name "white"))
    (terminal-print 0 (1- height) (format nil "H: ~d M: ~d"
                                          (slot-value player 'health)
                                          (slot-value player 'mana)))))

(defun display-lose ()
  (terminal-clear)
  (terminal-print 0 0 "You lose!")
  (terminal-print 0 1 "Press escape key to exit.")
  (terminal-refresh)
  (loop as key = (terminal-read)
        while (not (eq key bearlibterminal-ffi:+tk-escape+))))

(defun display-win ()
  (terminal-clear)
  (terminal-print 0 0 "You win!")
  (terminal-print 0 1 "Press escape key to exit.")
  (terminal-refresh)
  (loop as key = (terminal-read)
        while (not (eq key bearlibterminal-ffi:+tk-escape+))))

(defun enemy-p (thing)
  (typep thing 'enemy))

(defun no-enemies-left-p (map)
  (not (find-if #'enemy-p (alexandria:hash-table-values (game-map-mobiles map)))))

(defun main-loop ()
  (loop with player = (make-instance 'player-character)
        with map = (initialize-map (make-instance 'game-map) player)
        with viewport = (create-viewport map)
        with prev-key = nil
        as key = (terminal-read) do
          (when (handle-input key map viewport player prev-key)
            (setf prev-key key)
            (unless (or (= key bearlibterminal-ffi:+tk-r+)
                        (= key bearlibterminal-ffi:+tk-m+))
              (end-of-turn-tasks player map))
            (terminal-clear)
            (write-map-to-terminal player map viewport)
            (write-status-to-terminal player prev-key)
            (terminal-refresh))
        until (or (= key bearlibterminal-ffi:+tk-close+)
                  (= key bearlibterminal-ffi:+tk-escape+)
                  (eq (slot-value player 'location) nil)
                  (no-enemies-left-p map))
        finally (cond
                  ((eq (slot-value player 'location) nil) (display-lose))
                  ((no-enemies-left-p map) (display-win)))))

(defun start ()
  (terminal-open)
  (terminal-set "window: size=99x40, title='Crypts and Corpses'")
  (terminal-set (format nil "font: ~SDejaVuSansMono.ttf, size=16"
                        (namestring (asdf:system-relative-pathname :crypts-and-corpses "fonts/"))))
  (intro-message)
  (main-loop)
  (terminal-close))
